/*
 * builder_busroute.nut
 * This file is part of AroAI
 *
 * Copyright (C) 2010 - Charles Pigott (Lord Aro)
 *
 * AroAI is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * AroAI is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AroAI; if not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA  02110-1301  USA
 */


class Builder_BusRoute
	{
	function Main();
	function GetTowns();
	function BuildRoad(town_a, town_b);
	function BuildBusStop(town);
	function BuildRoadDepot(town);
	
	town_list = null;
	town_a = null;
	town_b = null;
	err = null;
	busStation_a = 0;
	busStation_b = 0;
	depot_tile_a = 0;
	depot_tile_b = 0;
	engine = 0;
	numToRemove = 1;
	}

function Builder_BusRoute::Main()
	{
		Debugs.Info("Planning on building bus route"); //TODO: make this more efficent
		if(AIGameSettings.IsDisabledVehicleType(AIVehicle.VT_ROAD))
		{
			Debugs.Error("ROAD VEHICLES ARE DISABLED. THIS VERSION OF AROAI ONLY USES ROAD VEHICLES");
			Debugs.Error("PLEASE RE-ENABLE THEM, THEN RESTART GAME");
			AroAI.Stop();
		}
		if(AIController.GetSetting("enable_road_vehs") == 0)
		{
			Debugs.Error("ROADS VEHICLE TYPE IS DISABLED. THIS VERSION OF AROAI ONLY USES ROAD VEHICLES");
			Debugs.Error("PLEASE RE-ENABLE THEM, THEN RESTART GAME");
			AroAI.Stop();
		}
		
		local towns = Builder_BusRoute.GetTowns();
		if(towns == null)
		{
			return;
		}
		local route = Builder_BusRoute.BuildRoad(town_a, town_b);
//		Debugs.Debug("route = " + route);
		if(route == null)
			return;
		
		local town = town_a;
		busStation_a = Builder_BusRoute.BuildBusStop(town);
		if(busStation_a == null)
			//TODO: Deal with all returns
			return;
		
		depot_tile_a = Builder_BusRoute.BuildRoadDepot(town_a);
		if(depot_tile_a == null)
			return;
		
		town = town_b;
		busStation_b = Builder_BusRoute.BuildBusStop(town);
		if(busStation_b == null)
			return;
		
		depot_tile_b = Builder_BusRoute.BuildRoadDepot(town_b);
		if(depot_tile_b == null)
			return;
		
		local bbe = VehicleManager.BuildBusEngines(depot_tile_a, busStation_a, busStation_b);
		if(bbe == null)
			return;

		bbe = VehicleManager.BuildBusEngines(depot_tile_b, busStation_b, busStation_a);
		if(bbe == null)
			return;
	}

function Builder_BusRoute::GetTowns()
	{
		AIRoad.SetCurrentRoadType(AIRoad.ROADTYPE_ROAD);
		town_list = AITownList();
		town_list.Valuate(AITown.GetPopulation);
		town_list.Sort(AIAbstractList.SORT_BY_VALUE, false);
		town_list.RemoveTop(numToRemove);
		town_a = town_list.Begin();
		town_list.RemoveTop(1);
		local tile_a = AITown.GetLocation(town_a);
//		town_list.Clear();
//		town_list = AITownList();
		town_list.Valuate(AITown.GetDistanceManhattanToTile, tile_a);
		town_list.KeepBelowValue(125);
		town_list.Valuate(AITown.GetRating, AICompany.COMPANY_SELF);
		town_list.KeepValue(AITown.TOWN_RATING_NONE);
		if(town_list.IsEmpty())
		{
			Debugs.Info("No serviceable towns within radius of " + town_a + ". Moving to next town");
			numToRemove++;
				return null;
		}	
		town_list.Valuate(AITown.GetPopulation);
		town_list.Sort(AIAbstractList.SORT_BY_VALUE, false);
		town_b = town_list.Begin(); //pick the top one
		      	return town_a, town_b;
	}

function Builder_BusRoute::BuildRoad(town_a, town_b)
	{
		Debugs.Info("Planning route from " + AITown.GetName(town_a) + " to " + AITown.GetName(town_b));
		local pathfinder = RoadPathFinder();
		pathfinder.cost.turn = 1;
		pathfinder.InitializePath([AITown.GetLocation(town_b)], [AITown.GetLocation(town_a)]);
		local path = false;
		local counter = 0;
		while (path == false)
		{
			path = pathfinder.FindPath(100);//100 attempts at finding path TODO: check
			counter++;
			AIController.Sleep(1);
		}
		if (path == null) //No path was found
		{
			Debugs.Warning("No route found");
				return null;
		}
		Debugs.Info("Route found. (Tried " + counter + " times) Building started");
		while (path != null) //If a path found, build a road over it
		{
			local par = path.GetParent();
			if (par != null)
			{
				local last_node = path.GetTile();
				if (AIMap.DistanceManhattan(path.GetTile(), par.GetTile()) == 1 )
				{
					if (!AIRoad.BuildRoad(path.GetTile(), par.GetTile()))
					{
						switch(AIError.GetLastError())
						{
							case AIError.ERR_VEHICLE_IN_THE_WAY:
								Debugs.Warning("Building road failed temporarily - vehicle in the way");
								// try again till the stupid vehicle moved?
								while(!AIRoad.BuildRoad(path.GetTile(), par.GetTile()));
							break;
							case AIRoad.ERR_ROAD_WORKS_IN_PROGRESS:
								Debugs.Warning("Building road failed - road works in progress");
								// good, someone else is building this, skip tile
							break;
							case AIError.ERR_ALREADY_BUILT:
								// even better; someone else already built this; silent ignore
							break;
							case AIError.ERR_LAND_SLOPED_WRONG:
								Debugs.Error("Building road failed - wrong slope");
								// TODO terraform
							case AIError.ERR_AREA_NOT_CLEAR:
								Debugs.Warning("Building road failed: not clear, demolishing tile..");
								local sign = AISign.BuildSign(path.GetTile(), "Clearing tile");
								// demolish and retry
								AITile.DemolishTile(path.GetTile())
								AIRoad.BuildRoad(path.GetTile(), par.GetTile());
								AISign.RemoveSign(sign);
							break;
							case AIRoad.ERR_ROAD_ONE_WAY_ROADS_CANNOT_HAVE_JUNCTIONS:
								Debugs.Error("Building road failed - no junctions on one way roads allowed");
								// hopefully can't happen? nobody will ever use those...
									return null;
							case AIError.ERR_NOT_ENOUGH_CASH:
								Debugs.Warning("Not enough money. Waiting for more");
								AIController.Sleep(500); //wait then retry
								AIRoad.BuildRoad(path.GetTile(), par.GetTile());
							break;
							default:
								Debugs.Warning("Unknown error during road building " + AIError.GetLastErrorString());
								AISign.BuildSign(path.GetTile(), AIError.GetLastErrorString());
									return null;
							break;	// no clue what this is; silent ignore for now
						}
					}
				}
				else
				{
					/* Build a bridge or tunnel. */
					if (!AIBridge.IsBridgeTile(path.GetTile()) && !AITunnel.IsTunnelTile(path.GetTile()))
					{
						/* If it was a road tile, demolish it first. Do this to work around expended roadbits. */
						if (AIRoad.IsRoadTile(path.GetTile())) AITile.DemolishTile(path.GetTile());
						if (AITunnel.GetOtherTunnelEnd(path.GetTile()) == par.GetTile())
						{
							if (!AITunnel.BuildTunnel(AIVehicle.VT_ROAD, path.GetTile()))
							{
								/* An error occured while building a tunnel. TODO: handle it. */
							}
						} 
						else
						{
							local bridge_list = AIBridgeList_Length(AIMap.DistanceManhattan(path.GetTile(), par.GetTile()) + 1);
							bridge_list.Valuate(AIBridge.GetMaxSpeed);
							bridge_list.Sort(AIAbstractList.SORT_BY_VALUE, false);
							if (!AIBridge.BuildBridge(AIVehicle.VT_ROAD, bridge_list.Begin(), path.GetTile(), par.GetTile()))
							{
								/* An error occured while building a bridge. TODO: handle it. */
							}
						}
					}
				}
			}
			path = par;
		}
		Debugs.Info("Road building finished");
		Debugs.Info(AITown.GetName(town_a) + " is now connected to " + AITown.GetName(town_b))
			return true;
	}

function Builder_BusRoute::BuildBusStop(town)
	{
		Debugs.Info("Building bus stop in " + AITown.GetName(town));
		// Find empty square as close to town centre as possible
		local spotFound = false;
		local range = 1;
		local max_range = sqrt(AITown.GetPopulation(town)/100) + 2; //TODO check value correctness 
		local area = AITileList();
		
		while (range < max_range)
		{
			area.AddRectangle(AITown.GetLocation(town) - AIMap.GetTileIndex(range, range), AITown.GetLocation(town) + AIMap.GetTileIndex(range, range));
			area.Valuate(AIRoad.IsRoadTile);
			area.KeepValue(1);
			area.Valuate(AIRoad.IsDriveThroughRoadStationTile);
			area.KeepValue(0);
			area.Valuate(AITile.GetSlope);
			area.KeepValue(AITile.SLOPE_FLAT);
			area.Valuate(AIRoad.GetNeighbourRoadCount);
			area.KeepValue(2);	// entrance and exit; allow 1 as well?
			if (area.Count())
			{
				for (local station = area.Begin(); !area.IsEnd(); station = area.Next())
				{
					//Debugs.Debug("station(line258) = " + station);
					local opening = getRoadTile(station);
					if(opening)
					{
						local buildStation = AIRoad.BuildDriveThroughRoadStation(station, opening, AIRoad.ROADVEHTYPE_BUS, AIStation.STATION_JOIN_ADJACENT);
						if(!buildStation)
						{
							switch(AIError.GetLastError())
							{
							case AIError.ERR_VEHICLE_IN_THE_WAY:
								Debugs.Warning("Bus stop building temporarily stopped - vehicle in the way");
								continue;
							break;
							case AIError.ERR_NOT_ENOUGH_CASH:
								Debugs.Info("Not enough money. Waiting for more");
								while(AICompany.GetBankBalance(AICompany.COMPANY_SELF) < AIRoad.GetBuildCost(AIRoad.ROADTYPE_ROAD, AIRoad.BT_BUS_STOP))
								{
									AIController.Sleep(50);
								}
								AIRoad.BuildDriveThroughRoadStation(station, opening, AIRoad.ROADVEHTYPE_BUS, AIStation.STATION_JOIN_ADJACENT);
								Debugs.Info("Bus stop successfully built");
								town_list.RemoveValue(town);
									return station;
							break;
							default:
								Debugs.Warning("Error building bus stop: " + AIError.GetLastErrorString() + " Trying again");
									continue;
							break;
							}
						}
						else
						{
						Debugs.Info("Bus stop successfully built");
						//Debugs.Debug("station(line294) = " + station);
						town_list.RemoveValue(town);
							return station;
						}
					}
				}
				range++;
			}
			else
			{
				range++;
			}
		}
		Debugs.Warning("No possible bus stop location found");
		return null;
	}

function Builder_BusRoute::BuildRoadDepot(town)
	{
		Debugs.Info("Building depot in " + AITown.GetName(town));
		local depotList = AIDepotList(AITile.TRANSPORT_ROAD);
		depotList.Valuate(AITile.GetClosestTown);
		depotList.KeepValue(town);
//		Debugs.Debug("depotList has " + depotList.Count() + " depots in it");
		//if depot already in town, use that one
		if(!depotList.IsEmpty())
		{
			local depot_tile = depotList.Begin();
			return depot_tile;
		}
		else
		{
			// Find empty square as close to station as possible
			local range = 1;
			local area = AITileList();
			local townLocation = AITown.GetLocation(town);
	
			while (range < 15)
			{
//				Debugs.Warning("DEBUG: looking with range " + range);
				area.AddRectangle(townLocation - AIMap.GetTileIndex(range, range), townLocation + AIMap.GetTileIndex(range, range));
//				Debugs.Warning("DEBUG: At line 246 there are " + area.Count() + " possible options");
				area.Valuate(AITile.IsBuildable);
				area.KeepValue(1);
//				Debugs.Warning("DEBUG: At line 248 there are " + area.Count() + " possible options");
				if (area.Count())
				{
					for (local depot_tile = area.Begin(); !area.IsEnd(); depot_tile = area.Next())
					{
						local depot_front = getRoadTile(depot_tile);
						if(depot_front)
						{
//							Debugs.Warning("DEBUG: tile " + depot_tile + " has opening " + depot_front);
							if (AIRoad.BuildRoad(depot_tile, depot_front) && AIRoad.BuildRoadDepot(depot_tile, depot_front))
							{
								Debugs.Info("Depot successfully built");
//								Debugs.Warning("DEBUG: (line261) depot_tile = " + depot_tile);
								return depot_tile;
							}
						}
					}
					range++; // the found options had no road connections; enlarge search area
				}
				else
				{
					range++;
					area.Clear;
				}
			}
			Debugs.Warning("Depot building in " + AITown.GetName(town) + " failed");
				return null;
		}
	}

function Builder_BusRoute::getRoadTile(tile) //from otvi
	{
		local adjacent = AITileList();
		adjacent.AddTile(tile - AIMap.GetTileIndex(1,0));
		adjacent.AddTile(tile - AIMap.GetTileIndex(0,1));
		adjacent.AddTile(tile - AIMap.GetTileIndex(-1,0));
		adjacent.AddTile(tile - AIMap.GetTileIndex(0,-1));
		adjacent.Valuate(AIRoad.IsRoadTile);
		adjacent.KeepValue(1);
		adjacent.Valuate(AIRoad.IsRoadStationTile);
		adjacent.KeepValue(0);
		adjacent.Valuate(AITile.GetSlope);
		adjacent.KeepValue(AITile.SLOPE_FLAT);
		if (adjacent.Count())
			return adjacent.Begin();
		else
			return null;
	}
