/*
 * main.nut
 * This file is part of AroAI
 *
 * Copyright (C) 2010 - Charles Pigott (Lord Aro)
 *
 * AroAI is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * AroAI is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AroAI; if not, see <http://www.gnu.org/licenses/> or
 * write to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
BEFORE RELEASE CHECK:
	run stress test
	make sure versions are same (main.nut, info.nut & makefile)
	run makefile

MINOR TODO: (x.x.x++)
	extra debugs

MAIN TODO: (x.x++.x)
	add configurable no. of buses per town
	add check for towns being pre-connected (see wiki)
	redo bus-station-in-town checker
	autoreplace
	reform debugs output
	remove failed bus stops (and depots)
	predict cost of route (and act according to answer)
	re-write town-finder (currently ignoring towns that have ben built through)
	make stations build properly adjacent (maybe)
	events manager
	better vehicle+cargo selector (think NoCAB)
	save/load support

WISHFUL THINKING TODO: (x++.x.x)
	Write own road pathfinder
	Cargo support
	Rail support
		Double rail support
	Air support
	ship support

	6.x.x by end :)
*/
import("pathfinder.road", "RoadPathFinder", 3);
import("pathfinder.rail", "RailPathFinder", 1);
require("debugs.nut");
require("vehiclemanager.nut");
require("builder_busroute.nut");
require("moneymanager.nut");

class AroAI extends AIController
{
	HQtown = 0;
	aiversion = "1.0.0(r52)";
	
	Builder_BusRoute = Builder_BusRoute();
	VehicleManager = VehicleManager();
	Manager = Manager();
}


function AroAI::Start()
	{
		this.Sleep(1);
		GetVersionsAndStuff();
		SetCompany(); //Set company stuff
		while (true) //Keep running. If Start() exits, the AI dies
		{
			this.Sleep(1);
			Debugs.Warning("Main loop started");
			Manager.ManageLoan();
			local vehList = AIVehicleList();
			vehList.Valuate(AIVehicle.GetVehicleType);
			vehList.KeepValue(AIVehicle.VT_ROAD);
			local numOfVehs = vehList.Count();
			if(AIGameSettings.GetValue("vehicle.max_roadveh") >  numOfVehs)
			{
				Builder_BusRoute.Main();
			}
			else
			{
			Debugs.Info("Max amount of road vehicles reached");
			this.Sleep(500);
			}
		}
	}

function AroAI::Stop()
	{
		Debugs.Error("Something gone wrong. Clearing all signs");
		AroAI.ClearAllSigns();
		Debugs.Error("Stopped");
		Debugs.Error("(The error is on purpose)");
		local var = 1/0;
 	}

function AroAI::Save()
	{
		Debugs.Warning("TODO: Add Save/Load functionality");
		local table = {}; //TODO: Add save data to the table.
			return table;
	}

function AroAI::Load(version, data)
	{
		//TODO: Add loading routines.
	}

function AroAI::SetCompany()
	{
		local a = AIBase.RandRange(15);
		if (0==a)   {AICompany.SetName("Arioa International");}
		if (1==a)   {AICompany.SetName("AroAI");}
		if (2==a)   {AICompany.SetName("Aro Transport");}
		if (3==a)   {AICompany.SetName("Aro & Co.");}
		if (4==a)   {AICompany.SetName("Aro");}
		if (5==a)   {AICompany.SetName("Aro Distribution");}
		if (6==a)   {AICompany.SetName("Aro Logistics");}
		if (7==a)   {AICompany.SetName("Aro Federal Delivery");}
		if (8==a)   {AICompany.SetName("Aro Ltd.");}
		if (9==a)   {AICompany.SetName("Aro Delivery");}
		if (10==a)  {AICompany.SetName("Aro Network");}
		if (11==a)  {AICompany.SetName("Aro Trans");}
		if (12==a)  {AICompany.SetName("Aro Services");}
		if (13==a)  {AICompany.SetName("Aro Management");}
		if (14==a)  {AICompany.SetName("Aro Constructions");}
		Debugs.Info(AICompany.GetName(AICompany.COMPANY_SELF) + " inaugurated");

		AICompany.SetPresidentGender(0);
		AICompany.SetPresidentName("Lord Aro");
		Debugs.Info(AICompany.GetPresidentName(AICompany.COMPANY_SELF) + " is the new president");

		AICompany.SetAutoRenewStatus(true);
		AICompany.SetAutoRenewMonths(-3);
		AICompany.SetAutoRenewMoney(0);
		BuildHQ();
	}

function AroAI::GetVersionsAndStuff()
	{
		Debugs.Info("AroAI v" + aiversion + " by Charles Pigott (Lord Aro) started");
		Debugs.Info("Special thanks to those who helped with the many problems had when making the AI")
		local version = GetVersion();
		Debugs.Info("Currently playing on OpenTTD version " + ((version & (15 << 28)) >> 28) + "." +
		((version & (15 << 24)) >> 24) + "." + ((version & (15 << 20)) >> 20) + "" + 
		(((version & (1 << 19)) >> 19)?" stable release, ":" r") + ((version & ((1 << 18) - 1))));
		AILog.Info("=======================================")
	}

function AroAI::BuildHQ() //from Rondje
	{
		if(AIMap.IsValidTile(AICompany.GetCompanyHQ(AICompany.COMPANY_SELF))) //from simpleai
			return;

		//Find second biggest town for HQ, just to be different :p
		local towns = AITownList();
		towns.Valuate(AITown.GetPopulation);
		towns.Sort(AIAbstractList.SORT_BY_VALUE, false);
		if(towns.Count == 1)
		{
			HQtown = towns.Begin();
		}
		else
		{
			towns.RemoveTop(1);
			HQtown = towns.Begin();
		}

		//Find empty 2x2 square as close to town centre as possible
		local maxRange = Sqrt(AITown.GetPopulation(HQtown)/100) + 5;
		local HQArea = AITileList();

		HQArea.AddRectangle(AITown.GetLocation(HQtown) - AIMap.GetTileIndex(maxRange, maxRange), AITown.GetLocation(HQtown) + AIMap.GetTileIndex(maxRange, maxRange));
		HQArea.Valuate(AITile.IsBuildableRectangle, 2, 2);
		HQArea.KeepValue(1);
		HQArea.Valuate(AIMap.DistanceManhattan, AITown.GetLocation(HQtown));
		HQArea.Sort(AIList.SORT_BY_VALUE, true);
		for (local tile = HQArea.Begin(); !HQArea.IsEnd(); tile = HQArea.Next()) 
		{
			if (AICompany.BuildCompanyHQ(tile)) 
			{
				AISign.BuildSign(tile, "AroAI HQ");
				Debugs.Info("HQ building completed");
					return;
			}
		}
		Debugs.Warning("No possible HQ location found");
	}

function AroAI::Sqrt(i) //from Rondje
	{
		if (i == 0)
			return 0;   // Avoid divide by zero
		local n = (i / 2) + 1;       // Initial estimate, never low
		local n1 = (n + (i / n)) / 2;
		while (n1 < n)
		{
			n = n1;
			n1 = (n + (i / n)) / 2;
		}
			return n;
	}

function AroAI::ClearAllSigns()
	{
		local i = 0;
		local sign_list = AISignList();
		for(i = 0; i < sign_list.Count(); ++i)
		{
			if(i != null && AISign.IsValidSign(i))
			{
				AISign.RemoveSign(i);
			}
		}
	}
